import torch
from torch import autograd

from Inpainting_zoo.utilities import *
from Inpainting_zoo.model import *

def calc_gradient_penalty(netD, real_data, fake_data, device):
    batch_size = real_data.size(0)
    alpha = torch.rand(batch_size, 1, 1, 1)
    alpha = alpha.expand_as(real_data)
    alpha = alpha.to(device)

    interpolates = alpha * real_data + (1 - alpha) * fake_data
    interpolates = interpolates.requires_grad_().clone()

    disc_interpolates = netD(interpolates.float())
    grad_outputs = torch.ones(disc_interpolates.size())
    grad_outputs = grad_outputs.to(device)

    gradients = autograd.grad(outputs=disc_interpolates, inputs=interpolates,
                                grad_outputs=grad_outputs, create_graph=True,
                                retain_graph=True, only_inputs=True)[0]

    gradients = gradients.view(batch_size, -1)
    gradient_penalty = ((gradients.norm(2, dim=1) - 1) ** 2).mean()

    return gradient_penalty


import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms
import torch.optim as optim
from torch.utils.data import DataLoader
from torch import autograd
import gc

import time
import pyprind

import matplotlib.pyplot as plt

from Inpainting_zoo.model import *
from Inpainting_zoo.loss import *
from Inpainting_zoo.utilities import *

### Parameters ###

batch_size = 6
start_epochs = 0
total_epochs = 8 
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

### Initializing the Model ###

net_gen = Generator(use_cuda=True)
net_local_dis = LocalDis()
net_global_dis = GlobalDis()

net_gen.to(device)
net_local_dis.to(device)
net_global_dis.to(device)

d_params = list(net_local_dis.parameters()) + list(net_global_dis.parameters())
optimizer_g = optim.Adam(net_gen.parameters(), lr=0.0001, betas=(0.5, 0.9))
optimizer_d = optim.Adam(d_params, lr=0.0001, betas=(0.5, 0.9))

criterionL1 = nn.L1Loss().to(device)

### Train Functions ###

def dis_forward(netD, ground_truth, x_inpaint):
    #assert ground_truth.size() == x_inpaint.size()
    batch_size = ground_truth.size(0)
    batch_data = torch.cat([ground_truth, x_inpaint], dim=0)
    batch_output = netD(batch_data.float())
    real_pred, fake_pred = torch.split(batch_output, batch_size, dim=0)

    return real_pred, fake_pred

def train_Inpainting(net_gen, net_local_dis, net_global_dis, dataset, optimizer_g, optimizer_d, criterionL1):

    epoch_loss = {'l1':0, 'ae':0, 'wgan_g':0, 'wgan_d':0, 'wgan_gp':0, 'g':0, 'd':0}
    train_loss = []

    epoch_start = time.time()
    gc.collect()
    torch.cuda.empty_cache()

    net_gen.train()
    net_local_dis.train()
    net_global_dis.train()

    bar = pyprind.ProgBar(len(dataset), bar_char='█')
    for idx, ground_truth in enumerate(dataset, 1):

        gc.collect()
        torch.cuda.empty_cache()

        batch_size = ground_truth.size(0)
        bboxes = random_bbox(batch_size=batch_size)
        x, mask = mask_image(ground_truth, bboxes)

        ground_truth = ground_truth.to(device)
        x = x.to(device)
        mask = mask.to(device)

        losses = {}

        ##################
        ### Prediction ###
        ##################

        x1, x2, offset_flow = net_gen(x.float(), mask)
        local_patch_gt = local_patch(ground_truth, bboxes)
        x1_inpaint = x1 * mask + x * (1. - mask)
        x2_inpaint = x2 * mask + x * (1. - mask)
        local_patch_x1_inpaint = local_patch(x1_inpaint, bboxes)
        local_patch_x2_inpaint = local_patch(x2_inpaint, bboxes)

        gc.collect()
        torch.cuda.empty_cache()

        ##########################
        ### Discriminator Loss ###
        ##########################

        ### Local Discriminator ###
                                                    
        local_patch_real_pred, local_patch_fake_pred = dis_forward(net_local_dis, local_patch_gt, local_patch_x2_inpaint.detach())

        gc.collect()
        torch.cuda.empty_cache()

        ### Global Discriminator ###

        global_real_pred, global_fake_pred = dis_forward(net_global_dis, ground_truth, x2_inpaint.detach())

        gc.collect()
        torch.cuda.empty_cache()

        ### Computing Losses ###

        losses['wgan_d'] = torch.mean(local_patch_fake_pred-local_patch_real_pred) + torch.mean(global_fake_pred-global_real_pred)
        
        local_penalty = calc_gradient_penalty(net_local_dis, local_patch_gt, local_patch_x2_inpaint.detach(), device)
        global_penalty = calc_gradient_penalty(net_global_dis, ground_truth, x2_inpaint.detach(), device)
        losses['wgan_gp'] = local_penalty + global_penalty

        gc.collect()
        torch.cuda.empty_cache()

        ######################
        ### Generator Loss ###
        ######################

        sd_mask = spatial_discounting_mask(use_cuda=True)

        losses['l1'] = 1.2*criterionL1(local_patch_x1_inpaint*sd_mask, local_patch_gt*sd_mask) + criterionL1(local_patch_x2_inpaint*sd_mask, local_patch_gt*sd_mask)

        losses['ae'] = 1.2*criterionL1(x1*(1.-mask), ground_truth*(1.-mask)) + criterionL1(x2*(1.-mask), ground_truth*(1.-mask))

        local_patch_real_pred_gen, local_patch_fake_pred_gen = dis_forward(net_local_dis, local_patch_gt, local_patch_x2_inpaint)
        global_real_pred_gen, global_fake_pred_gen = dis_forward(net_global_dis, ground_truth, x2_inpaint)
        losses['wgan_g'] = - torch.mean(local_patch_fake_pred_gen) - torch.mean(global_fake_pred_gen)

        gc.collect()
        torch.cuda.empty_cache()

        ####################
        ### Forward Pass ###
        ####################

        for k in losses.keys():
            if not losses[k].dim() == 0:
                losses[k] = torch.mean(losses[k])

        #####################
        ### Backward Pass ###
        #####################
        with torch.autograd.set_detect_anomaly(True):
            if idx%5 !=0:
                optimizer_d.zero_grad()
                losses['d'] = losses['wgan_d'] + losses['wgan_gp']*10
                losses['d'].backward()
                optimizer_d.step()
            else:
                optimizer_g.zero_grad()
                losses['g'] = losses['l1']*1.2 + losses['ae']*1.2 + losses['wgan_g']*0.001
                losses['g'].backward()
                optimizer_g.step()

        gc.collect()
        torch.cuda.empty_cache()

        #####################
        ### Visualization ###
        #####################

        for key in losses.keys():
            epoch_loss[key] += losses[key].item()/len(dataset)

        #train_loss.append(losses)
        
        bar.update()
        gc.collect()
        torch.cuda.empty_cache()

    gc.collect()
    torch.cuda.empty_cache()

    print(" | Train Loss: Generator: {0}  |  Disctiminator: {1}".format(epoch_loss['g'], epoch_loss['d']))

def val_Inpainting(net_gen, dataset,criterionL1):

    epoch_loss = {'l1':0, 'ae':0, 'wgan_g':0, 'g':0}
    eval_loss = []

    gc.collect()
    torch.cuda.empty_cache()

    net_gen.eval()
    net_local_dis.eval()
    net_global_dis.eval()

    with torch.no_grad():
        bar = pyprind.ProgBar(len(dataset), bar_char='█')
        for idx, ground_truth in enumerate(dataset, 1):

            gc.collect()
            torch.cuda.empty_cache()

            batch_size = ground_truth.size(0)
            bboxes = random_bbox(batch_size=batch_size)
            x, mask = mask_image(ground_truth, bboxes)

            ground_truth = ground_truth.to(device)
            x = x.to(device)
            mask = mask.to(device)

            losses = {}

            ##################
            ### Prediction ###
            ##################

            x1, x2, offset_flow = net_gen(x.float(), mask)
            local_patch_gt = local_patch(ground_truth, bboxes)
            x1_inpaint = x1 * mask + x * (1. - mask)
            x2_inpaint = x2 * mask + x * (1. - mask)
            local_patch_x1_inpaint = local_patch(x1_inpaint, bboxes)
            local_patch_x2_inpaint = local_patch(x2_inpaint, bboxes)

            gc.collect()
            torch.cuda.empty_cache()

            ######################
            ### Generator Loss ###
            ######################

            sd_mask = spatial_discounting_mask(use_cuda=True)

            losses['l1'] = 1.2*criterionL1(local_patch_x1_inpaint*sd_mask, local_patch_gt*sd_mask) + criterionL1(local_patch_x2_inpaint*sd_mask, local_patch_gt*sd_mask)

            losses['ae'] = 1.2*criterionL1(x1*(1.-mask), ground_truth*(1.-mask)) + criterionL1(x2*(1.-mask), ground_truth*(1.-mask))

            local_patch_real_pred, local_patch_fake_pred = dis_forward(net_local_dis, local_patch_gt, local_patch_x2_inpaint)
            global_real_pred, global_fake_pred = dis_forward(net_global_dis, ground_truth, x2_inpaint)
            losses['wgan_g'] = - torch.mean(local_patch_fake_pred) - torch.mean(global_fake_pred)

            gc.collect()
            torch.cuda.empty_cache()

            ####################
            ### Forward Pass ###
            ####################

            losses['g'] = losses['l1']*1.2 + losses['ae']*1.2 + losses['wgan_g']*0.001
            for k in losses.keys():
                if not losses[k].dim() == 0:
                    losses[k] = torch.mean(losses[k])

            #####################
            ### Visualization ###
            #####################

            for key in losses.keys():
                epoch_loss[key] += losses[key].item()/len(dataset)

            #eval_loss.append(losses)
            
            bar.update()
            gc.collect()
            torch.cuda.empty_cache()

    gc.collect()
    torch.cuda.empty_cache()
    print(" | Validation Loss: Generator: {0}".format(epoch_loss['g']))


### Training ###

train_loss = []
val_loss = []

# for epoch in range(start_epochs+1, total_epochs+start_epochs+1):
#     print("Starting Epoch[{0}/{1}]".format(epoch, total_epochs+start_epochs))
    
#     epoch_start = time.time()

#     train_epoch_loss, _ = train(net_gen, net_local_dis, net_global_dis, trainloader, optimizer_g, optimizer_d, criterionL1)
#     train_loss.append(train_epoch_loss)
#     print(" | Train Loss: Generator: {0}  |  Disctiminator: {1}".format(train_epoch_loss['g'], train_epoch_loss['d']))

#     val_epoch_loss, _ = evaluate(net_gen, valloader, criterionL1)
#     val_loss.append(val_epoch_loss)
#     print(" | Validation Loss: Generator: {0}".format(val_epoch_loss['g']))
    

#     epoch_end = time.time()

#     minutes, seconds = epoch_time(epoch_end, epoch_start)

#     print("Finished Epoch[{0}/{1}]".format(epoch, total_epochs+start_epochs))


import torch
import torch.nn as nn
import torchvision
import numpy as np
import torch.nn.functional as F
import math
from torch.autograd import Variable
import torch.utils.model_zoo as model_zoo
from UNet_zoo.model import *
from UNet_zoo.train import *
from UNetPP_zoo.model import *
from UNetPP_zoo.train import *
from PointNet_zoo.model import *
from PointNet_zoo.train import *
from StackGAN_zoo.model import *
from StackGAN_zoo.train1 import *
from StackGAN_zoo.train2 import *
from CharCNN_zoo.model import *
from PSPNet_zoo.model import *
from PSPNet_zoo.train import *
from Inpainting_zoo.model import *
from Inpainting_zoo.train import *
from AdapAttnIC_zoo.model import *

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
def models(name = None):
    if name is None:
        return "No Models loaded"
    else:
        return f"{name} model loaded successfully"

class UNet:
    def model(num_channels = 1, pretrained = False):
        if pretrained == True:
            model = U_Net(num_channels)
            url = 'https://github.com/Vinayak-VG/ModelZoo_PyPi/releases/download/Weights/U-Net_Image_Seg_Norm.pt'
            checkpoint = model_zoo.load_url(url, map_location=device)
            model.load_state_dict(checkpoint['model_state_dict'])
            print("Pre-Trained Model Loaded Succesfully")
        else:
            model = U_Net(num_channels)
            print("Model Loaded Succesfully") 
        return model
    
    def train(model, train_dataloader, optimizer, loss, device):
        train_UNet(model, train_dataloader, optimizer, loss, device)
    
    def val(model, val_dataloader, loss, device):
        eval_UNet(model, val_dataloader, loss, device)

class UNetPP:
    def model(num_channels = 1, pretrained = False):
        if pretrained == True:
            model = UNet_PP(num_channels = num_channels, denseblock = DenseBlock, basicdownblock = BasicDownBlock, basicupblock = BasicUpBlock)
            url = "https://github.com/Vinayak-VG/ModelZoo_PyPi/releases/download/Weights/U-Net++.pt"    
            checkpoint = model_zoo.load_url(url, map_location=device)
            model.load_state_dict(checkpoint['model_state_dict'])
            print("Pre-Trained Model Loaded Succesfully")
        else:    
            model = UNet_PP(num_channels, DenseBlock, BasicDownBlock, BasicUpBlock)
            print("Model Loaded Succesfully") 
        return model
    
    def train(model, train_dataloader, optimizer, loss, device):
        train_UNetPP(model, train_dataloader, optimizer, loss, device)
    
    def val(model, val_dataloader, loss, device):
        eval_UNetPP(model, val_dataloader, loss, device)

class PointNet:
    def model(pretrained = False):
        if pretrained == True:
            model = Point_Net(T_Net)
            url = "https://github.com/Vinayak-VG/ModelZoo_PyPi/releases/download/Weights/PointNet_Best.pt"
            checkpoint = model_zoo.load_url(url, map_location=device)
            model.load_state_dict(checkpoint['model_state_dict'])
            print("Pre-Trained Model Loaded Succesfully")
        else:
            model = Point_Net(T_Net)
            model = model.apply(weights_init)
            print("Model Loaded Succesfully") 
        return model
    
    def train(model, train_dataloader, optimizer, device):
        train_PointNet(model, train_dataloader, optimizer, device)
    
    def val(model, val_dataloader, device):
        val_PointNet(model, val_dataloader, device)

class CharCNN:
    def model(pretrained = False):
        if pretrained == True:
            model = Char_CNN()
            print("Pre-Trained Model Loaded Succesfully") 
            return model
        else:
            model = Char_CNN()
            print("Model Loaded Succesfully") 
            return model

class StackGAN:
    def model(name = "None", pretrained = False):
        if pretrained == True:
            if name == "None":
                return "Please specify the name of Model"
            elif name == 'StageI_Gen':
                model = StageI_GAN_Gen(Conditioning_Augmentation_StageI)
                url = "https://github.com/Vinayak-VG/ModelZoo_PyPi/releases/download/Weights/StageI_Gen_GAN_GPT2-3.pt"
                checkpoint = model_zoo.load_url(url, map_location=device)
                model.load_state_dict(checkpoint['model_state_dict'])
                print("Pre-Trained Model Loaded Succesfully")
                return model
            elif name == 'StageI_Dis':
                model = StageI_GAN_Dis(DownSample1)
                url = "https://github.com/Vinayak-VG/ModelZoo_PyPi/releases/download/Weights/StageI_Dis_GAN_GPT2.pt"
                checkpoint = model_zoo.load_url(url, map_location=device)
                model.load_state_dict(checkpoint['model_state_dict'])
                print("Pre-Trained Model Loaded Succesfully")
                return model
            elif name == 'StageII_Gen':
                model = StageII_GAN_Gen(DownSample2, ResidualBlock, UpSampling2, Conditioning_Augmentation_StageII)
                url = "https://github.com/Vinayak-VG/ModelZoo_PyPi/releases/download/Weights/StageII_Gen_GAN_GPT2.pt"
                checkpoint = model_zoo.load_url(url, map_location=device)
                model.load_state_dict(checkpoint['model_state_dict'])
                print("Pre-Trained Model Loaded Succesfully")
                return model
            elif name == 'StageII_Dis':
                model = StageII_GAN_Dis(DownSample3)
                url = "https://github.com/Vinayak-VG/ModelZoo_PyPi/releases/download/Weights/StageII_Dis_GAN_GPT2.pt"
                checkpoint = model_zoo.load_url(url, map_location=device)
                model.load_state_dict(checkpoint['model_state_dict'])
                print("Pre-Trained Model Loaded Succesfully")
                return model
        else:
            if name == "None":
                return "Please specify the name of Model"
            elif name == 'StageI_Gen':
                model = StageI_GAN_Gen(Conditioning_Augmentation_StageI)
                model = model.apply(weights_init)
                print("Model Loaded Succesfully") 
                return model
            elif name == 'StageI_Dis':
                model = StageI_GAN_Dis(DownSample1)
                model = model.apply(weights_init)
                print("Model Loaded Succesfully") 
                return model
            elif name == 'StageII_Gen':
                model = StageII_GAN_Gen(DownSample2, ResidualBlock, UpSampling2, Conditioning_Augmentation_StageII)
                model = model.apply(weights_init)
                print("Model Loaded Succesfully") 
                return model
            elif name == 'StageII_Dis':
                model = StageII_GAN_Dis(DownSample3)
                model = model.apply(weights_init)
                print("Model Loaded Succesfully") 
                return model

    def trainI(dataset, optimizerD1, optimizerG1, device):
        train_StackGAN_StageI(dataset, optimizerD1, optimizerG1, device)

    def trainII(dataset, optimizerD1, optimizerG1, device):
        train_StackGAN_StageII(dataset, optimizerD1, optimizerG1, device)

class Inpainting:
    def model(name = "None", pretrained = False):
        if pretrained == True:
            if name == "None":
                return "Please specify the name of Model"
            elif name == 'Generator':
                model = Generator(use_cuda=True)
                print("Pre-Trained Model Loaded Succesfully")
                return model
            elif name == 'LocalDis':
                model = LocalDis()
                print("Pre-Trained Model Loaded Succesfully")
                return model
            elif name == 'GlobalDis':
                model = GlobalDis()
                print("Pre-Trained Model Loaded Succesfully")
                return model
        else:
            if name == "None":
                return "Please specify the name of Model"
            elif name == 'Generator':
                model = Generator(use_cuda=True)
                print("Model Loaded Succesfully") 
                return model
            elif name == 'LocalDis':
                model = LocalDis()
                print("Model Loaded Succesfully") 
                return model
            elif name == 'GlobalDis':
                model = GlobalDis()
                print("Model Loaded Succesfully") 
                return model

    def train(net_gen, net_local_dis, net_global_dis, dataset, optimizer_g, optimizer_d, criterionL1):
        train_Inpainting(net_gen, net_local_dis, net_global_dis, dataset, optimizer_g, optimizer_d, criterionL1)

    def val(net_gen, dataset,criterionL1):
        val_Inpainting(net_gen, dataset,criterionL1)

class PSPNet:
    def model(pretrained = False):
        if pretrained == True:
            model = PSP_Net()
            print("Pre-Trained Model Loaded Succesfully") 
            return model
        else: 
            model = PSP_Net()
            print("Model Loaded Succesfully") 
            return model

    def train(model, dataset, optimizer, epoch, device):
        train_PSPNet(model, dataset, optimizer, epoch, device)
    
    def val(model, dataset, epoch, device):
        val_PSPNet(model, dataset, epoch, device)

class AdapAttnIC:
    def model(embed_dim, hidden_dim, feature_dim, vocab_size, weights_matrix, name = 'None',pretrained = False):
        if pretrained == True:
            if name == "None":
                return "Please specify the name of Model"
            elif name == 'Encoder':
                model = Encoder()
                print("Pre-Trained Model Loaded Succesfully")
                return model
            elif name == 'Decoder':
                model = Decoder(embed_dim, hidden_dim, feature_dim, vocab_size, weights_matrix, device)
                print("Pre-Trained Model Loaded Succesfully")
                return model
        if pretrained == False:
            if name == "None":
                return "Please specify the name of Model"
            elif name == 'Encoder':
                model = Encoder()
                print("Model Loaded Succesfully") 
                return model
            elif name == 'Decoder':
                model = Decoder(embed_dim, hidden_dim, feature_dim, vocab_size, weights_matrix, device)
                print("Model Loaded Succesfully") 
                return model






import torch
import torch.nn as nn
import torchvision
import torchvision.transforms as transforms
import torch.nn.functional as F

import os
from torchvision.utils import save_image
import glob
import math 
from sklearn.model_selection import train_test_split
import gc
import time
import torch.optim as optim
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from UNetPP_zoo.model import *
from tqdm import trange

epoch_train_losses = []              # Defining an empty list to store the epoch losses
epoch_val_losses = []             
accu_train_epoch = []                # Defining an empty list to store the accuracy per epoch
accu_val_epoch = []

model = UNet_PP(num_channels = 1, denseblock = DenseBlock, basicdownblock = BasicDownBlock, basicupblock = BasicUpBlock)

def make_tensor(tensor):
      if torch.cuda.is_available():
        return torch.cuda.FloatTensor(tensor)
      else:
        return torch.FloatTensor(tensor)

class BCEDiceLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, input, target):
        bce = F.binary_cross_entropy(input, target)
        smooth = 1e-5
        num = target.size(0)
        input = input.view(num, -1)
        target = target.view(num, -1)
        intersection = (input * target)
        dice = (2. * intersection.sum(1) + smooth) / (input.sum(1) + target.sum(1) + smooth)
        dice = 1 - dice.sum() / num
        return 0.5 * bce + dice             # We are giving more weightage to Dice Loss

batch_size = 1
epochs = 10
lr = 1e-3
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
loss = BCEDiceLoss()
model.to(device)
optimizer = torch.optim.Adam(model.parameters(), lr)


def iou_score(output, target):
    smooth = 1e-5

    if torch.is_tensor(output):
        output = output.data.cpu().numpy()
    if torch.is_tensor(target):
        target = target.data.cpu().numpy()
    output_ = output > 0.5
    target_ = target > 0.5
    intersection = (output_ & target_).sum()
    union = (output_ | target_).sum()

    return (intersection + smooth) / (union + smooth)


def train_UNetPP(model, dataset, optimizer, criterion, device):

    train_loss_batch = []
    accu_train_batch = []
    model.train()
    for idx,(images, labels) in enumerate(dataset):
        images = images.to(device)
        labels = labels.to(device)

        #Forward Pass
        output = model(make_tensor(images))
        output = torch.clip(output, 0.0025, 0.9975)            # I am clipping the output because if it becomes 0 or 1 then there is a chance that loss function can explode
        labels = torch.round(labels)
        train_loss = criterion(output, labels)
        train_loss_batch.append(train_loss)
        output = torch.round(output)
        acc = iou_score(output, labels)
        accu_train_batch.append(acc)
        print(f"Batch: {idx + 1}   Train Loss: {train_loss:.5f}   Accuracy: {acc:.5f}")
        # Backward
        optimizer.zero_grad()
        train_loss.backward()
        optimizer.step()
    epoch_train_losses.append(sum(train_loss_batch)/len(dataset))
    accu_train_epoch.append(sum(accu_train_batch)/len(dataset))
    print(f"Train Epoch Loss: {(sum(train_loss_batch)/len(dataset)):.5f}   Train Epoch Accuracy: {(sum(accu_train_batch)/len(dataset)):.5f}")

def eval_UNetPP(model, dataset, criterion, device):

    val_loss_batch = []
    accu_val_batch = []
    model.eval()
    for idx,(images, labels) in enumerate(dataset):
        with torch.no_grad():
            images = images.to(device)
            labels = labels.to(device)
            #Forward Pass
            output = model(make_tensor(images))
            torch.clip(output, 0.0025, 0.9975)
            # Loss
            val_loss = criterion(output, labels)
            val_loss_batch.append(val_loss)
            output = torch.round(output)
            acc = iou_score(output, labels)
            accu_val_batch.append(acc)
    epoch_val_losses.append((sum(val_loss_batch))/len(dataset))
    accu_val_epoch.append((sum(accu_val_batch))/len(dataset))
    print(f"Val Epoch Loss: {((sum(val_loss_batch))/len(dataset)):.5f}   Val Epoch Accuracy: {((sum(accu_val_batch))/len(dataset)):.5f}")


import os
import torch
import numpy as np
import torch.nn.functional as F


def same_padding(images, ksizes, strides, rates):
    assert len(images.size()) == 4
    batch_size, channel, rows, cols = images.size()
    out_rows = (rows + strides[0] - 1) // strides[0]
    out_cols = (cols + strides[1] - 1) // strides[1]
    effective_k_row = (ksizes[0] - 1) * rates[0] + 1
    effective_k_col = (ksizes[1] - 1) * rates[1] + 1
    padding_rows = max(0, (out_rows-1)*strides[0]+effective_k_row-rows)
    padding_cols = max(0, (out_cols-1)*strides[1]+effective_k_col-cols)
    # Pad the input
    padding_top = int(padding_rows / 2.)
    padding_left = int(padding_cols / 2.)
    padding_bottom = padding_rows - padding_top
    padding_right = padding_cols - padding_left
    paddings = (padding_left, padding_right, padding_top, padding_bottom)
    images = torch.nn.ZeroPad2d(paddings)(images)
    return images

def extract_image_patches(images, ksizes, strides, rates, padding='same'):

    assert len(images.size()) == 4
    assert padding in ['same', 'valid']
    batch_size, channel, height, width = images.size()

    if padding == 'same':
        images = same_padding(images, ksizes, strides, rates)
    elif padding == 'valid':
        pass
    else:
        raise NotImplementedError('Unsupported padding type: {}.\
                Only "same" or "valid" are supported.'.format(padding))

    unfold = torch.nn.Unfold(kernel_size=ksizes,
                             dilation=rates,
                             padding=0,
                             stride=strides)
    patches = unfold(images)
    return patches  # [N, C*k*k, L], L is the total number of such blocks

def reduce_mean(x, axis=None, keepdim=False):
    if not axis:
        axis = range(len(x.shape))
    for i in sorted(axis, reverse=True):
        x = torch.mean(x, dim=i, keepdim=keepdim)
    return x


def reduce_std(x, axis=None, keepdim=False):
    if not axis:
        axis = range(len(x.shape))
    for i in sorted(axis, reverse=True):
        x = torch.std(x, dim=i, keepdim=keepdim)
    return x


def reduce_sum(x, axis=None, keepdim=False):
    if not axis:
        axis = range(len(x.shape))
    for i in sorted(axis, reverse=True):
        x = torch.sum(x, dim=i, keepdim=keepdim)
    return x

def make_color_wheel():
    RY, YG, GC, CB, BM, MR = (15, 6, 4, 11, 13, 6)
    ncols = RY + YG + GC + CB + BM + MR
    colorwheel = np.zeros([ncols, 3])
    col = 0
    # RY
    colorwheel[0:RY, 0] = 255
    colorwheel[0:RY, 1] = np.transpose(np.floor(255 * np.arange(0, RY) / RY))
    col += RY
    # YG
    colorwheel[col:col + YG, 0] = 255 - np.transpose(np.floor(255 * np.arange(0, YG) / YG))
    colorwheel[col:col + YG, 1] = 255
    col += YG
    # GC
    colorwheel[col:col + GC, 1] = 255
    colorwheel[col:col + GC, 2] = np.transpose(np.floor(255 * np.arange(0, GC) / GC))
    col += GC
    # CB
    colorwheel[col:col + CB, 1] = 255 - np.transpose(np.floor(255 * np.arange(0, CB) / CB))
    colorwheel[col:col + CB, 2] = 255
    col += CB
    # BM
    colorwheel[col:col + BM, 2] = 255
    colorwheel[col:col + BM, 0] = np.transpose(np.floor(255 * np.arange(0, BM) / BM))
    col += + BM
    # MR
    colorwheel[col:col + MR, 2] = 255 - np.transpose(np.floor(255 * np.arange(0, MR) / MR))
    colorwheel[col:col + MR, 0] = 255
    return colorwheel

def compute_color(u, v):
    h, w = u.shape
    img = np.zeros([h, w, 3])
    nanIdx = np.isnan(u) | np.isnan(v)
    u[nanIdx] = 0
    v[nanIdx] = 0
    colorwheel = make_color_wheel()
    ncols = np.size(colorwheel, 0)
    rad = np.sqrt(u ** 2 + v ** 2)
    a = np.arctan2(-v, -u) / np.pi
    fk = (a + 1) / 2 * (ncols - 1) + 1
    k0 = np.floor(fk).astype(int)
    k1 = k0 + 1
    k1[k1 == ncols + 1] = 1
    f = fk - k0
    for i in range(np.size(colorwheel, 1)):
        tmp = colorwheel[:, i]
        col0 = tmp[k0 - 1] / 255
        col1 = tmp[k1 - 1] / 255
        col = (1 - f) * col0 + f * col1
        idx = rad <= 1
        col[idx] = 1 - rad[idx] * (1 - col[idx])
        notidx = np.logical_not(idx)
        col[notidx] *= 0.75
        img[:, :, i] = np.uint8(np.floor(255 * col * (1 - nanIdx)))
    return img

def flow_to_image(flow):
    out = []
    maxu = -999.
    maxv = -999.
    minu = 999.
    minv = 999.
    maxrad = -1
    for i in range(flow.shape[0]):
        u = flow[i, :, :, 0]
        v = flow[i, :, :, 1]
        idxunknow = (abs(u) > 1e7) | (abs(v) > 1e7)
        u[idxunknow] = 0
        v[idxunknow] = 0
        maxu = max(maxu, np.max(u))
        minu = min(minu, np.min(u))
        maxv = max(maxv, np.max(v))
        minv = min(minv, np.min(v))
        rad = np.sqrt(u ** 2 + v ** 2)
        maxrad = max(maxrad, np.max(rad))
        u = u / (maxrad + np.finfo(float).eps)
        v = v / (maxrad + np.finfo(float).eps)
        img = compute_color(u, v)
        out.append(img)
    return np.float32(np.uint8(out))

def random_bbox(batch_size, image_shape=(256,256,3), mask_shape=(128, 128), margin=(0,0), mask_batch_same=True):
    img_height, img_width, _ = image_shape
    h, w = mask_shape
    margin_height, margin_width = margin
    maxt = img_height - margin_height - h
    maxl = img_width - margin_width - w
    bbox_list = []
    if mask_batch_same:
        t = np.random.randint(margin_height, maxt)
        l = np.random.randint(margin_width, maxl)
        bbox_list.append((t, l, h, w))
        bbox_list = bbox_list * batch_size
    else:
        for i in range(batch_size):
            t = np.random.randint(margin_height, maxt)
            l = np.random.randint(margin_width, maxl)
            bbox_list.append((t, l, h, w))

    return torch.tensor(bbox_list, dtype=torch.int64)

def bbox2mask(bboxes, height, width, max_delta_h, max_delta_w):
    batch_size = bboxes.size(0)
    mask = torch.zeros((batch_size, 1, height, width), dtype=torch.float32)
    for i in range(batch_size):
        bbox = bboxes[i]
        delta_h = np.random.randint(max_delta_h // 2 + 1)
        delta_w = np.random.randint(max_delta_w // 2 + 1)
        mask[i, :, bbox[0] + delta_h:bbox[0] + bbox[2] - delta_h, bbox[1] + delta_w:bbox[1] + bbox[3] - delta_w] = 1.
    return mask

def mask_image(x, bboxes, image_shape=(256,256,3), max_delta_shape=(32,32), mask_type='hole'):
    height, width, _ = image_shape
    max_delta_h, max_delta_w = max_delta_shape
    mask = bbox2mask(bboxes, height, width, max_delta_h, max_delta_w)
    if x.is_cuda:
        mask = mask.cuda()

    if mask_type == 'hole':
        result = x * (1. - mask)
    elif mask_type == 'mosaic':
        # TODO: Matching the mosaic patch size and the mask size
        mosaic_unit_size = 12
        downsampled_image = F.interpolate(x, scale_factor=1. / mosaic_unit_size, mode='nearest')
        upsampled_image = F.interpolate(downsampled_image, size=(height, width), mode='nearest')
        result = upsampled_image * mask + x * (1. - mask)
    else:
        raise NotImplementedError('Not implemented mask type.')

    return result, mask

def spatial_discounting_mask(spatial_discounting_gamma=0.9, mask_shape=(128, 128), discounted_mask=True, use_cuda=False):
    gamma = spatial_discounting_gamma
    height, width = mask_shape
    shape = [1, 1, height, width]
    if discounted_mask:
        mask_values = np.ones((height, width))
        for i in range(height):
            for j in range(width):
                mask_values[i, j] = max(
                    gamma ** min(i, height - i),
                    gamma ** min(j, width - j))
        mask_values = np.expand_dims(mask_values, 0)
        mask_values = np.expand_dims(mask_values, 0)
    else:
        mask_values = np.ones(shape)
    spatial_discounting_mask_tensor = torch.tensor(mask_values, dtype=torch.float32)
    if use_cuda:
        spatial_discounting_mask_tensor = spatial_discounting_mask_tensor.cuda()
    return spatial_discounting_mask_tensor

def local_patch(x, bbox_list):
    assert len(x.size()) == 4
    patches = []
    for i, bbox in enumerate(bbox_list):
        t, l, h, w = bbox
        patches.append(x[i, :, t:t + h, l:l + w])
    return torch.stack(patches, dim=0)

def epoch_time(epoch_end, epoch_start):
    epoch_length = epoch_end - epoch_start

    minutes = epoch_length//60
    seconds = epoch_length - minutes*60

    return (minutes, seconds)


import os, glob
import gc
import time
import numpy as np
import matplotlib.pyplot as plt

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
import torch.optim as optim

from PSPNet_zoo.resnet import resnet101, resnet50
from PSPNet_zoo.model import *

### Parameters ###

batch_size = 16
start_epochs = 0
epochs = 2
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
PATH = "/content/drive/MyDrive/Projects/Clubs/Analytics/Coord Projects/Model Zoo/PSPNet/datasets/ADEChallengeData2016"
CHECKPOINT = "/content/drive/MyDrive/Projects/Clubs/Analytics/Coord Projects/Model Zoo/PSPNet/checkpoints"

### Initializing the Model ###

model = PSP_Net()
model.to(device)

criterion = nn.BCEWithLogitsLoss().to(device)
optimizer = optim.Adam(model.parameters(), lr=0.001)

if os.path.exists(os.path.join(CHECKPOINT, "model.pth")):
    checkpoints = torch.load(os.path.join(CHECKPOINT, "model.pth"))

    model.load_state_dict(checkpoints['model_state_dict'])
    optimizer.load_state_dict(checkpoints['optimizer_state_dict'])
    start_epochs = checkpoints['epoch']

### Training ###

train_loss = []
train_acc = []
val_loss = []
val_acc = []

def plot_loss(epoch, train_loss, val_loss):
    epochs = np.arange(epoch+1)
    plt.plot(epochs, train_loss, label="Train Loss", color="green")
    plt.plot(epochs, val_loss, label="Validation Loss", color="red")

    plt.title("Training Loss - "+str(epoch))
    plt.xlabel("Epochs")
    plt.ylabel("Loss")
    plt.legend(loc="best")

    plt.savefig(os.path.join(CHECKPOINT, "Loss.png"))
    plt.close()

def train_PSPNet(model, dataset, optimizer, epoch, device):

    gc.collect()
    torch.cuda.empty_cache()

    model.train()
    epoch_train_loss = []
    for idx, batch in enumerate(dataset, 1):
        time_batch_start = time.time()
        image = batch['image'].to(device)
        label = batch['label'].to(device)
        preds = model(image)
        loss = criterion(preds, label)
        
        loss.backward()
        optimizer.step()
        epoch_train_loss.append(loss.item())
        time_batch_end = time.time()
        print("Epoch[{0}]: Batch[{1}]   Train Loss: {2}     Time: {3}".format(epoch, idx, loss.item(), time_batch_end-time_batch_start))
        gc.collect()
        torch.cuda.empty_cache()

    train_loss.append(epoch_train_loss[-1])

    gc.collect()
    torch.cuda.empty_cache()

def val_PSPNet(model, dataset, epoch, device):

    model.eval()
    epoch_val_loss = []
    with torch.no_grad():
        for idx, batch in enumerate(dataset, 1):
            time_batch_start = time.time()
            image = batch['image'].to(device)
            label = batch['label'].to(device)
            preds = model(image)
            loss = criterion(preds, label)

            epoch_val_loss.append(loss.item())
            time_batch_end = time.time()
            print("Epoch[{0}]: Batch[{1}]   Val Loss: {2}     Time: {3}".format(epoch, idx, loss.item(), time_batch_end-time_batch_start))

        val_loss.append(epoch_val_loss[-1])

    gc.collect()
    torch.cuda.empty_cache()

    


import torch
import torch.nn as nn
import torchvision
import torchvision.transforms as transforms
import torch.nn.functional as F
import os
import json
import glob 
import gc
import time
import torch.optim as optim
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
import numpy as np
import matplotlib.pyplot as plt
from torch.autograd import Variable
from PointNet_zoo.model import *

epoch_train_losses = []  
epoch_test_losses = []
epoch_val_losses = []              
accu_train_epoch = []
accu_test_epoch = []
accu_val_epoch = []

batch_size = 32
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

model = Point_Net(T_Net).to(device)
model = model.apply(weights_init)

epochs = 20
lr = 0.001
model = model.to(device)
optimizer = torch.optim.Adam(model.parameters(), lr, eps=1e-08, weight_decay= 1e-4)

def feature_transform_regularizer(trans):
    d = trans.size()[1]
    batchsize = trans.size()[0]
    I = torch.eye(d)[None, :, :]
    if trans.is_cuda:
        I = I.cuda()
    loss = torch.mean(torch.norm(torch.bmm(trans, trans.transpose(2,1)) - I, dim=(1,2)))
    return loss

def criterion(pred, target, trans_feat):
        loss = F.nll_loss(pred, target)
        mat_diff_loss = feature_transform_regularizer(trans_feat)
        total_loss = loss + mat_diff_loss * 0.001
        return total_loss

def accuracy(y_pred, y):
    _, predicted = torch.max(y_pred, 1)
    correct = (predicted == y).sum().item()
    return correct/(batch_size*2048)

def make_tensor(tensor):
  if torch.cuda.is_available():
    return torch.cuda.FloatTensor(tensor)
  else:
    return torch.FloatTensor(tensor)

def train_PointNet(model, dataset, optimizer, device):

    train_loss_batch = []
    accu_train_batch = []
    model.train()
    for idx,(data, labels, seg) in enumerate(dataset):
        optimizer.zero_grad()
        data = data.to(device)
        labels = labels.to(device)
        seg = seg.to(device)
        #Forward Pass
        data = data.float()
        labels = labels.long()
        seg = seg.long()
        data = data.transpose(2, 1)

        output, A = model(data, to_categorical(labels, 16))
        A = A.to(device)
        output = output.contiguous().view(-1, 50)
        seg = seg.view(-1, 1)[:, 0]
        pred_choice = output.data.max(1)[1]
        correct = pred_choice.eq(seg.data).cpu().sum()
        accu_train_batch.append(correct.item() / (batch_size * 2048))
        train_loss = criterion(output, seg, A)
        train_loss_batch.append(train_loss.item())
        # Backward
        train_loss.backward()
        optimizer.step()
    epoch_train_losses.append(sum(train_loss_batch)/len(dataset))
    accu_train_epoch.append(sum(accu_train_batch)/len(dataset))
    print(f"Train Epoch Loss: {sum(train_loss_batch)/len(dataset):.5f}   Train Epoch Accuracy: {sum(accu_train_batch)/len(dataset)*100:.5f}")

def val_PointNet(model, dataset, device):
    
    val_loss_batch = []
    accu_val_batch = []
    model.eval()
    for idx,(data, labels, seg) in enumerate(dataset):
      with torch.no_grad():
        data = data.to(device)
        labels = labels.to(device)
        seg = seg.to(device)
        #Forward Pass
        data = data.float()
        labels = labels.long()
        seg = seg.long()
        data = data.transpose(2, 1)

        output, A = model(data, to_categorical(labels, 16))
        A = A.to(device)
        output = output.contiguous().view(-1, 50)
        seg = seg.view(-1, 1)[:, 0]
        pred_choice = output.data.max(1)[1]
        correct = pred_choice.eq(seg.data).cpu().sum()
        accu_val_batch.append(correct.item() / (batch_size * 2048))
        val_loss = criterion(output, seg, A)
        val_loss_batch.append(val_loss.item())
    epoch_val_losses.append((sum(val_loss_batch))/len(dataset))
    accu_val_epoch.append((sum(accu_val_batch))/len(dataset))
    print(f"Val Epoch Loss: {(sum(val_loss_batch))/len(dataset):.5f} Val Epoch Accuracy: {(sum(accu_val_batch))/len(dataset)*100:.5f}")


import torch
import torch.nn as nn
import torchvision
import torchvision.transforms as transforms
import torch.nn.functional as F

import os
from torchvision.utils import save_image
import glob
import math 
from sklearn.model_selection import train_test_split
import gc
import time
import torch.optim as optim
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from UNet_zoo.model import *
from tqdm import trange

epoch_train_losses = []              # Defining an empty list to store the epoch losses
epoch_val_losses = []             
accu_train_epoch = []                # Defining an empty list to store the accuracy per epoch
accu_val_epoch = []

model = U_Net()

def make_tensor(tensor):
      if torch.cuda.is_available():
        return torch.cuda.FloatTensor(tensor)
      else:
        return torch.FloatTensor(tensor)

batch_size = 1
epochs = 1
lr = 0.001
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

loss = nn.BCEWithLogitsLoss()                                         
model = model.to(device)
optimizer = torch.optim.SGD(model.parameters(), lr, momentum = 0.99)

def iou_score(output, target):
    smooth = 1e-5

    if torch.is_tensor(output):
        output = output.data.cpu().numpy()
    if torch.is_tensor(target):
        target = target.data.cpu().numpy()
    output_ = output > 0.5
    target_ = target > 0.5
    intersection = (output_ & target_).sum()
    union = (output_ | target_).sum()

    return (intersection + smooth) / (union + smooth)

def train_UNet(model, dataset, optimizer, criterion, device):

    train_loss_batch = []
    accu_train_batch = []
    model.train()
    for idx,(images, labels) in enumerate(dataset):
      images = images.to(device)
      labels = labels.to(device)

      #Forward Pass
      output = model(make_tensor(images))
      torch.clip(output, 0.0025, 0.9975)            # I am clipping the output because if it becomes 0 or 1 then there is a chance that loss function can explode
      labels = torch.round(labels)                           # Rounding of the image pixel values to 0 or 1 since some of them had values like 0.001, 0.998 etc
      train_loss = loss(output,labels)
      train_loss_batch.append(train_loss)
      output = torch.round(output)
      acc = iou_score(output, labels)
      accu_train_batch.append(acc)
      print(f"Batch: {idx + 1}   Train Loss: {train_loss}   Accuracy: {acc}")
      # Backward
      optimizer.zero_grad()
      train_loss.backward()
      optimizer.step()
    epoch_train_losses.append(sum(train_loss_batch)/len(dataset))
    accu_train_epoch.append(sum(accu_train_batch)/len(dataset))
    print(f"Train Epoch Loss: {sum(train_loss_batch)/len(dataset)}   Train Epoch Accuracy: {sum(accu_train_batch)/len(dataset)}")

def eval_UNet(model, dataset, criterion, device):

    val_loss_batch = []
    accu_val_batch = []
    model.eval()
    for idx,(images, labels) in enumerate(dataset):
      with torch.no_grad():
        images = images.to(device)
        labels = labels.to(device)
        #Forward Pass
        output = model(make_tensor(images))
        torch.clip(output, 0.0025, 0.9975)
        labels = torch.round(labels)
        # Loss
        val_loss = criterion(output,labels)
        val_loss_batch.append(val_loss)
        torch.round(output)
        acc = iou_score(output, labels)
        accu_val_batch.append(acc)
    epoch_val_losses.append((sum(val_loss_batch))/len(dataset))
    accu_val_epoch.append((sum(accu_val_batch))/len(dataset))
    print(f"Val Epoch Loss: {(sum(val_loss_batch))/len(dataset)}   Val Epoch Accuracy: {(sum(accu_val_batch))/len(dataset)}")


import pandas as pd
import numpy as np
import random
import os
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
import torchvision.transforms as transforms
import torch.optim as optim
from torchvision.utils import save_image
from PIL import Image
import PIL
import pickle
from torch.utils.data import Dataset
from glob import glob
import time
import gc
from torch.utils.data import DataLoader
from torch.utils.data import random_split
from torchvision.utils import make_grid
import matplotlib.pyplot as plt
from sentence_transformers import SentenceTransformer, util
from StackGAN_zoo.model import *


batch_size = 64
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

StageI_Gen = StageI_GAN_Gen(Conditioning_Augmentation_StageI).to(device)
StageI_Gen = StageI_Gen.apply(weights_init)
StageI_Dis = StageI_GAN_Dis(DownSample1).to(device)
StageI_Dis = StageI_Dis.apply(weights_init)
sbert_model = SentenceTransformer('paraphrase-mpnet-base-v2')

epoch_D1losses = []             
epoch_G1losses = []
epoch_D2losses = []             
epoch_G2losses = []
epoch_Real_Score = []
epoch_Fake_Score = []
epoch_Generator_Score = []


epochs = 600
lrG = 0.00002
lrD = 0.0002

optimizerD1 = torch.optim.Adam(StageI_Dis.parameters(), lr=lrD, betas=(0.5,0.999))
optimizerG1 = torch.optim.Adam(StageI_Gen.parameters(), lr=lrG, betas=(0.5,0.999))

BCEloss = nn.BCELoss()

def KL_loss(mu, logvar):
    # -0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
    KLD_element = mu.pow(2).add_(logvar.exp()).mul_(-1).add_(1).add_(logvar)
    KLD = torch.mean(KLD_element).mul_(-0.5)
    return KLD

def train_StageI_Dis(real_images, wrong_images, text, optimizer):

    optimizer.zero_grad()

    real_images = real_images.to(device)
    text = text.to(device)
    real_pred = StageI_Dis(real_images, text)
    real_targets = torch.ones(real_images.size(0),1)
    real_pred = real_pred.to(device)
    real_targets = real_targets.to(device)
    real_loss = BCEloss(real_pred, real_targets)
    real_score = torch.mean(real_pred).item()

    fake_images, mu, logvar = StageI_Gen(text)
    fake_pred1 = StageI_Dis(fake_images, text)
    fake_targets1 = torch.zeros(fake_images.size(0),1)
    fake_pred1 = fake_pred1.to(device)
    fake_targets1 = fake_targets1.to(device)
    fake_loss1 = BCEloss(fake_pred1, fake_targets1)
    fake_score1 = torch.mean(fake_pred1).item()

    wrong_images = wrong_images.to(device)
    fake_pred2 = StageI_Dis(wrong_images, text)
    fake_targets2 = torch.zeros(wrong_images.size(0),1)
    fake_pred2 = fake_pred2.to(device)
    fake_targets2 = fake_targets2.to(device)
    fake_loss2 = BCEloss(fake_pred2, fake_targets2)
    fake_score2 = torch.mean(fake_pred2).item()

    discriminator_loss = real_loss + (fake_loss1 + fake_loss2)/2
    discriminator_loss.backward()
    optimizer.step()

    return discriminator_loss.item(), real_score, (fake_score1+fake_score2)/2


def train_StageI_Gen(text, optimizer):

    optimizer.zero_grad()

    text = text.to(device)
    generator_images, mu, logvar = StageI_Gen(text)
    generator_pred = StageI_Dis(generator_images, text)
    generator_targets = torch.ones(batch_size, 1)
    generator_pred = generator_pred.to(device)
    generator_targets = generator_targets.to(device)
    gen_bin_loss = BCEloss(generator_pred, generator_targets)
    generator_score = torch.mean(generator_pred).item()
    kl_loss = KL_loss(mu, logvar)

    generator_loss = gen_bin_loss + 2*kl_loss
    generator_loss.backward()
    optimizer.step()

    return generator_loss.item(), gen_bin_loss, kl_loss, generator_score

def save_samples(index1, text, show=True):
    fake_images, a, b = StageI_Gen(text)
    fake_images = fake_images[0:16,:,:,:]
    fake_fname = 'generated-images-{}.png'.format(index1)
    save_image((fake_images), os.path.join("/content/drive/MyDrive/GAN Images/Birds/birds-5", fake_fname), nrow=4)
    if show:
        fig, ax = plt.subplots(figsize=(8, 8))
        ax.set_xticks([]); ax.set_yticks([])
        ax.imshow(make_grid(fake_images.cpu().detach(), nrow=8).permute(1, 2, 0))

''' Stage1 Training'''

def train_StackGAN_StageI(dataset, optimizerD1, optimizerG1, device):

    train_D1Loss_batch = []
    train_G1Loss_batch = []
    train_real_score = []
    train_fake_score = []
    train_generator_score = []

    for idx,(real_images, text) in enumerate(dataset):

        '''If you want to use HuggingFace Sentence Transformer'''
        text = list(text)
        embedding = []
        for i in range(len(text)):
            my_file = open(text[i], "r")
            content = my_file.read()
            embedding.append(content)
        emb = sbert_model.encode(embedding)
        emb = torch.from_numpy(emb)

        wrong_images = torch.flip(real_images, [0])
        discriminator_loss, real_score, fake_score = train_StageI_Dis(real_images, wrong_images, emb, optimizerD1)
        generator_loss, a, b, generator_score = train_StageI_Gen(emb, optimizerG1)
        train_D1Loss_batch.append(discriminator_loss)
        train_G1Loss_batch.append(generator_loss)
        train_real_score.append(real_score)
        train_fake_score.append(fake_score)
        train_generator_score.append(generator_score)

        torch.cuda.empty_cache()
        torch.cuda.reset_peak_memory_stats()
        torch.cuda.empty_cache()
        gc.collect()

    epoch_D1losses.append(sum(train_D1Loss_batch)/len(dataset))
    epoch_G1losses.append(sum(train_G1Loss_batch)/len(dataset))
    epoch_Real_Score.append(sum(train_real_score)/len(dataset))
    epoch_Fake_Score.append(sum(train_fake_score)/len(dataset))
    epoch_Generator_Score.append(sum(train_generator_score)/len(dataset))

    print(f"Discriminator Epoch Loss: {epoch_D1losses[-1]:.5f}   Generator Epoch Loss: {epoch_G1losses[-1]:.5f}   Real Score: {epoch_Real_Score[-1]:.5f}   Fake Score: {epoch_Fake_Score[-1]:.5f}   Generator Score: {epoch_Generator_Score[-1]:.5f}")

    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()
    torch.cuda.empty_cache()
    gc.collect()


